/* Utility class for holding collision-related information */
class Collision {
  PVector position, surfaceNormal;
  float t; /* Where in the time-step the collision occured; [0.00 - 1.00]. */
  GameEntity target = null; //What the ball collided with
  
  Collision() {
    position = new PVector(0, 0);
    surfaceNormal = new PVector(0, 0);
    t = 0;
  }
  
  Collision(float x, float y, float t){
    this();
    position.set(x, y, 0);
    this.t = t;
  }
  
  Collision(float x, float y, float normalX, float normalY, float t){
    this(x, y, t);
    this.surfaceNormal.set(normalX, normalY, 0);
  }
}

/*
Determine the intersection between a line segment and an axis-aligned bounding box.
If the line intersects the box in more than one place, the intersection point closest 
to the origin of the line will be returned.

Only detects intersections with lines coming from "outside" of the box. If the start point
of the line is located inside the box, no intersections will be detected.

Returns an instance of Collision, or NULL if the line doesn't touch the box.
*/
Collision intersectLineWithAABB(PLine segment, float xMin, float yMin, float xMax, float yMax){
  float collisionX = 0, collisionY = 0, collisionT = MAX_FLOAT;
  float normalX = 0, normalY = 0;
  float x, y, t;
  
  float deltaX = segment.end.x - segment.start.x;
  float deltaY = segment.end.y - segment.start.y;
  
  if ( deltaX > 0 ){
    //Left
    t = (xMin - segment.start.x) / deltaX;
    if ((t > 0) && (t <= 1) && (t < collisionT)) {
      //got a collision. Check if it's valid (within the Y bounds)
      y = segment.start.y + t * deltaY;
      if ( (y >= yMin) && (y <= yMax) ) {
        collisionX = xMin;
        collisionY = y;
        collisionT = t;
        normalX = -1; 
        normalY = 0;
      }
    }
  } else {
    //Right
    t = (xMax - segment.start.x) / deltaX;
    if ((t > 0) && (t <= 1) && (t < collisionT)) {
      //got a collision. Check if it's valid (within the Y bounds)
      y = segment.start.y + t * deltaY;
      if ( (y >= yMin) && (y <= yMax) ) {
        collisionX = xMax;
        collisionY = y;
        collisionT = t;
        normalX = 1; 
        normalY = 0;
      }
    }
  }
  
  if ( deltaY > 0 ) {
    //Top
    t = (yMin - segment.start.y) / deltaY;
    if ((t > 0) && (t <= 1) && (t < collisionT)) {
      //got a collision. Check if it's valid (within the Y bounds)
      x = segment.start.x + t * deltaX;
      if ( (x >= xMin) && (x <= xMax) ) {
        collisionX = x;
        collisionY = yMin;
        collisionT = t;
        normalX = 0; 
        normalY = -1;
      }
    }
  } else {
    //Bottom
    t = (yMax - segment.start.y) / deltaY;
    if ((t > 0) && (t <= 1) && (t < collisionT)) {
      //got a collision. Check if it's valid (within the Y bounds)
      x = segment.start.x + t * deltaX;
      if ( (x >= xMin) && (x <= xMax) ) {
        collisionX = x;
        collisionY = yMax;
        collisionT = t;
        normalX = 0; 
        normalY = 1;
      }
    }
  }
 
  if (collisionT != MAX_FLOAT){
    return new Collision(collisionX, collisionY, normalX, normalY, collisionT);
  } else {
    return null;
  }
}

/*
Determine the intersection between a line segment and an axis-aligned bounding box.
*/
Collision intersectLineWithAABB(PLine segment, AABB aabb){
  return intersectLineWithAABB(segment, aabb.xMin, aabb.yMin, aabb.xMax, aabb.yMax);
}

/*
Determine the intersection between a line segment and an axis-aligned bounding box
that has been "expanded" by X units in all directions. This can be used to perform 
a swept collision test between an axis-aligned, moving square and a static AABB.

Note that this is *not* the right way to detect collisions between a ball (i.e. circle) 
and an AABB, but it's a good-enough approximation for this simple game.
*/
Collision intersectLineWithExpandedAABB(PLine segment, AABB aabb, float expandBy){
  return intersectLineWithAABB(
    segment, 
    aabb.xMin - expandBy, 
    aabb.yMin - expandBy, 
    aabb.xMax + expandBy, 
    aabb.yMax + expandBy
  );
}
