//Basic "engine" that manages interaction and transitions 
//between various game states (e.g. "playing" and "paused").
//Also handles some pseudo-global resources like fonts.
class GameEngine {
  public PFont font;
  
  private Stack<GameState> states; //The stack of states
  String[] maps;      //List of available maps
  int currentMap = 0; //The current/most recently used map
  
  GameEngine(String[] maps){
    states = new Stack<GameState>();
    this.maps = maps;
  }
  
  void init(String title){
    if ( frame != null ){
      frame.setTitle(title);
    }
    
    smooth();
    background(255, 255, 255);
    noCursor();
    
    font = loadFont("AgencyFB-Reg-48.vlw");
    textFont(font);
  };
  
  void cleanup(){
    // cleanup the all states
    while ( !states.isEmpty() ) {
      states.peek().cleanup();
      states.pop();
    }
    println("GameEngine Cleanup\n");
  };

  void changeState(GameState state){
    // cleanup the current state
    if ( !states.isEmpty() ) {
      states.peek().cleanup();
      states.pop();
    }

    // store and init the new state
    states.push(state);
    states.peek().init(this);
  };
  
  void pushState(GameState state){
    // pause current state
    if ( !states.isEmpty() ) {
      states.peek().pause();
    }

    // store and init the new state
    states.push(state);
    states.peek().init(this);
  };
  
  void popState(){
    // cleanup the current state
    if ( !states.isEmpty() ) {
      states.peek().cleanup();
      states.pop();
    }

    // resume previous state
    if ( !states.isEmpty() ) {
      states.peek().resume();
    }
  };

  void handleEvents(){
    // let the state handle events
    states.peek().handleEvents(this);
  };
  
  void update(){
    // let the state update the game
    states.peek().update(this);
  };
  
  void draw(){
    // let the state draw the screen
    states.peek().draw(this);
  };
  
  //Pass mouse & keyboard events to the active state
  void mousePressed(){
    states.peek().mousePressed(this);
  };
  void mouseReleased(){
    states.peek().mouseReleased(this);
  };
  void keyPressed(){
    states.peek().keyPressed(this);
  };
  void keyReleased(){
    states.peek().keyReleased(this);
  };
  
  //Get the name of the first map
  String firstMap(){
    if ( maps.length > 0 ){
      currentMap = 0;
      return maps[currentMap];
    } else {
      return null;
    }
  }
  
  //Get the name of the next map
  String nextMap(){
    currentMap++;
    if ( currentMap < maps.length ){
      return maps[currentMap];
    } else {
      return null;
    }
  }  
}

//Base class for game states
abstract class GameState {
  protected void GameState(){}
  
  void init(GameEngine game){};
  void cleanup(){};

  void pause(){};
  void resume(){};
    
  void handleEvents(GameEngine game){};
  void update(GameEngine game){};
  void draw(GameEngine game){};
  
  void mousePressed(GameEngine game){};
  void mouseReleased(GameEngine game){};
  void keyPressed(GameEngine game){};
  void keyReleased(GameEngine game){};
}

//Base class for game entities like bricks, balls and the paddle
class GameEntity {
  float restitution = 1.00; //the "bounciness" of the object
  PlayState state;          //a convenience reference to the PlayState that "owns" the entity
  protected boolean dead = false; //dead entities are automatically removed at the end of the frame
  
  void handleEvents(){};
  void update(float deltaTime){};
  void draw(){};
  
  //Detect collisions between a ball's path and the entity.
  Collision checkForCollision(PLine path, float radius){
    return null;
  };
  void onCollision(Collision collision, Ball ball){};
  
  boolean dead(){
    return dead;
  };
}
