//The "play" state. This is where the bulk of the gameplay logic lies.
class PlayState extends GameState {
  String mapName; //Current map
  Arena arena;
  Paddle paddle;
  ArrayList<GameEntity> entities;      //Objects in play
  ArrayList<GameEntity> addedEntities; //A buffer for holding entities added in the current frame. This is necessary because 
                                       //we can't modify the main entity list while iterating over it. 
  int points = 0;
  int ballsInPlay = 0;
  int lives = 3;
  int remainingBricks = 0;
      
  long lastFrameTime; //nanoseconds
  float mapTime; //Time spent on the current map (in seconds)
  final float minPhysicsStep = 0.001;
  
  HashMap<String, AudioSample> sounds;
  boolean enableSound = true;
  
  PlayState(String mapName, int initialScore, int initialLives){
    this.mapName = mapName;
    points = initialScore;
    lives = max(initialLives, 0);
    
    entities = new ArrayList<GameEntity>(50);
    addedEntities = new ArrayList<GameEntity>(50);
  }
  
  PlayState(String mapName){
    this(mapName, 0, 3);
  }
  
  PlayState(){
    this("classic");
  }
  
  void init(GameEngine game){
    println("PlayState : init()");
    noCursor();
    
    //Create the game arena
    arena = new Arena((width - arenaWidth)/2, (height - arenaHeight)/2, arenaWidth, arenaHeight);
    addEntity(arena);
    
    //Create the paddle
    paddle = new Paddle();
    paddle.position.set(
      arena.width / 2,    //In the center of the arena
      (arena.height - paddle.height) - paddleOffset, //A little way off the bottom
      0
    );
    addEntity(paddle);
    
    //Create one ball
    addEntity(new Ball(arena.width / 2, arena.height / 2, 6, true));
    ballsInPlay = 1;
    
    //Load the map (i.e. bricks)
    loadMap(mapName, numBricksHorizontal, numBricksVertical, arenaWidth / numBricksHorizontal, brickHeight);
    
    //Move all of the above to the main entity list
    flushAddedEntities();
    
    //Load sound effects
    sounds = new HashMap<String, AudioSample>();
    sounds.put("hitbrick", minim.loadSample("hit.wav"));
    sounds.put("hitpaddle", minim.loadSample("bounce3.wav"));
    AudioSample hitWall = minim.loadSample("wall2.wav");
    if (hitWall.hasControl(Controller.GAIN)){
      hitWall.setGain(-6);
    }
    sounds.put("hitwall", hitWall);
    
    lastFrameTime = System.nanoTime();
    mapTime = 0;
  }
  
  void update(GameEngine game){
    //If the player has destroyed all bricks, declare victory
    if ( remainingBricks <= 0 ){
      game.changeState(new Victory(points, lives, mapTime));
      return;
    }
    
    //Check for lost balls / game over
    if ( ballsInPlay <= 0 ){
      if ( lives > 0 ){
        lives--;
        //Use one of the remaining balls
        //The ball gets created outside the map and automatically moves 
        //itself to the paddle when it gets updated for the first time.
        addEntity(new Ball(arena.width / 2, arena.height * 2, 6, true)); 
        ballsInPlay++;
      } else {
        //Game over
        game.pushState(new GameOver());
        return;
      }
    }
    
    //Calculate how much time has passed since the last frame
    long currentTime = System.nanoTime();
    float deltaTime = (currentTime - lastFrameTime) / 1000000000f; //Convert nanoseconds to seconds 
    if (deltaTime < minPhysicsStep){
      return; //No point in updating the physics too often, it would be imperceptible anyway
    }
    lastFrameTime = currentTime;
    mapTime += deltaTime;
    
    //Run the simulation
    for(GameEntity entity : entities){
      entity.update(deltaTime);
    }
    
    //Remove dead entities
    for(int i = entities.size()-1; i >= 0; i--){
      if ( entities.get(i).dead() ){
        entities.remove(i);
      }
    }
    
    //Add any newly created entities 
    flushAddedEntities();
  }
  
  void handleEvents(GameEngine game){
    for(GameEntity entity : entities){
      entity.handleEvents();
    }
  }
  
  void draw(GameEngine game){
    //Clear the screen
    background(255);
    
    //Draw everything
    translate(arena.left, arena.top);
    for(GameEntity entity : entities){
      entity.draw();
    }
    
    drawUI();
  }
  
  void drawUI(){
    textAlign(LEFT, BASELINE);
    
    //Display remaining lives/balls
    float radius = 7, spacing = 5;

    color fillColor = color(150, 240, 255);
    fill(fillColor);
    stroke(blendColor(fillColor, color(100), SUBTRACT));
    strokeWeight(2);
    
    for(int i = 0; i < lives; i++ ){
      ellipse(
        5 + i * (radius + spacing) + radius,
        arena.height - radius,
        radius,
        radius
      );
    }
    
    textFont(game.font, 24);
    textAlign(LEFT);
    fill(0);
    
    //Display FPS
    /*
    if ( frameCount > 30 ){
      text(round(frameRate) + " FPS", 220, arena.height);
    }
    //*/
    
    //Display score
    String score = nf(points, 5);
    text(score, arena.width - 55, arena.height);
  }
  
  void addPoints(int plusPoints){
    points += plusPoints;
  }
  
  void addEntity(GameEntity entity){
    entity.state = this;
    addedEntities.add(entity);
  }
  
  void flushAddedEntities(){
    if ( !addedEntities.isEmpty() ){
      entities.addAll(addedEntities);
      addedEntities.clear();
    }
  }
  
  boolean loadMap(String mapName, int numBricksHorizontal, int numBricksVertical, float brickWidth, float brickHeight){
    //Load the brick layout image
    PImage mapBricks = loadImage(mapName + ".png");
    if ( (mapBricks == null) || (mapBricks.width <= 0) || (mapBricks.height <= 0) ){
      return false;
    }
    
    //Load the image that contains brick metadata
    PImage mapMeta = loadImage(mapName + "-metadata.png");
    if ( (mapMeta == null) || (mapMeta.width <= 0) || (mapMeta.height <= 0) ){
      return false;
    }
    
    //Reset the number of destructible bricks on the map
    remainingBricks = 0;
    
    //If the map is smaller than the required dimensions it is centered. Otherwise, it is cropped to fit.
    int leftOffset = 0;
    if ( mapBricks.width < numBricksHorizontal ){
      leftOffset = (numBricksHorizontal - mapBricks.width) / 2;
    }
    int mapWidth = min(mapBricks.width, numBricksHorizontal);
    int mapHeight = min(mapBricks.height, numBricksVertical);
    
    //Convert each pixel to a brick. Pixels that are completely black or completely white are treated
    //as empty space.
    color black = color(0, 0, 0), white = color(255, 255, 255); 
    for ( int y = 0; y < mapHeight; y++ ){
      for ( int x = 0; x < mapWidth; x++ ){
        
          //Get the brick color
          color brickColor = mapBricks.get(x, y);
          if ( (brickColor == black) || (brickColor == white) ){
            continue;
          }
          brickColor = brickColor | 0xFF000000; //Bricks are always opaque
          
          //The number of points awarded for destroying this brick is stored
          //in the Red channel of the metadata image. To make it easier to "read"
          //the metadata visually, points are encoded as "red = 100 + pts*10"
          color meta = mapMeta.get(x, y);
          int brickPoints = round((red(meta) - 100) / 10); 
          
          //Create the brick
          Brick brick = new Brick(
            (x + leftOffset) * brickWidth, 
            y * brickHeight, 
            brickWidth, 
            brickHeight
          );
          brick.clr = brickColor;
          brick.points = brickPoints;
          //valuable bricks are more "bouncy"
          brick.restitution = constrain(map(brickPoints, 1, 7, 1.00, 1.03), 1.00, 1.10); 
          
          addEntity(brick);
          remainingBricks++;
      }
    }
    
    return true;
  }
  
  void ballLost(){
    ballsInPlay--;
  }
  
  void brickDestroyed(){
    remainingBricks--;
  }
  
  void playSound(String soundName){
    if ( !enableSound ){
      return;
    }
    AudioSample sound = sounds.get(soundName);
    if ( sound != null ){
      sound.trigger();
    }
  }
  
  void mousePressed(GameEngine game){
    //If the player clicks the left mouse button, unstick one ball from the paddle and
    //fire it off in a random direction.
    if ( mouseButton == LEFT ){
      for(GameEntity entity : entities){
        if ( entity instanceof Ball ){
          Ball ball = (Ball)entity;
          if ( ball.stickToPaddle ){ 
            ball.stickToPaddle = false;
            ball.velocity = rotateVector(new PVector(0, -initialBallVelocity), radians(random(-45, 45)));
            break;
          }
        }
      }
    }
  }
  
  void keyPressed(GameEngine game){
    //Pause the game when the "P" key is pressed
    if ( key == 'P' || key == 'p'){
      game.pushState(new PauseState());
    //Enable/disable sound effects with "M"
    } else if ( key == 'M' || key == 'm' ){
      enableSound = !enableSound;
    }
  }
  
  void cleanup(){
    entities.clear();
    addedEntities.clear();
    
    //The Minim library requires us to close() all sound effects once we're done with them.
    for(Map.Entry entry : sounds.entrySet()){
      AudioSample sound = (AudioSample)entry.getValue();
      sound.close();
    }
    
    println("PlayState : cleanup()");
  }
  
  void pause(){
    println("PlayState : pause()");
  }
  
  void resume(){
    noCursor();
    lastFrameTime = System.nanoTime(); 
    println("PlayState : resume()");
  }
}
